from dataclasses import dataclass
from enum import Enum
from typing import NamedTuple

from a7p import profedit_pb2

__all__ = ['A7PFactory']


class DistanceTable(Enum):
    SUBSONIC = (  # 25-400
        25, 50, 75, 100, 110, 120, 130, 140, 150, 155, 160, 165, 170, 175, 180, 185, 190, 195, 200, 205, 210, 215, 220,
        225, 230, 235, 240, 245, 250, 255, 260, 265, 270, 275, 280, 285, 290, 295, 300, 305, 310, 315, 320, 325, 330,
        335, 340, 345, 350, 355, 360, 365, 370, 375, 380, 385, 390, 395, 400,
    )
    LOW_RANGE = (  # 100-700
        100, 150, 200, 225, 250, 275, 300, 320, 340, 360, 380, 400, 410, 420, 430, 440,
        450, 460, 470, 480, 490, 500, 505, 510, 515, 520, 525, 530, 535, 540, 545, 550,
        555, 560, 565, 570, 575, 580, 585, 590, 595, 600, 605, 610, 615, 620, 625, 630,
        635, 640, 645, 650, 655, 660, 665, 670, 675, 680, 685, 690, 695, 700
    )
    MEDIUM_RANGE = (  # 100 - 1000
        100, 200, 250, 300, 325, 350, 375, 400, 420, 440, 460, 480, 500, 520, 540, 560, 580, 600, 610, 620, 630, 640,
        650, 660, 670, 680, 690, 700, 710, 720, 730, 740, 750, 760, 770, 780, 790, 800, 805, 810, 815, 820, 825, 830,
        835, 840, 845, 850, 855, 860, 865, 870, 875, 880, 885, 890, 895, 900, 905, 910, 915, 920, 925, 930, 935, 940,
        945, 950, 955, 960, 965, 970, 975, 980, 985, 990, 995, 1000
    )
    LONG_RANGE = (  # 100 - 1700
        100, 200, 250, 300, 350, 400, 420, 440, 460, 480, 500, 520, 540, 560, 580, 600, 610, 620, 630, 640, 650, 660,
        670, 680, 690, 700, 710, 720, 730, 740, 750, 760, 770, 780, 790, 800, 810, 820, 830, 840, 850, 860, 870, 880,
        890, 900, 910, 920, 930, 940, 950, 960, 970, 980, 990, 1000, 1005, 1010, 1015, 1020, 1025, 1030, 1035, 1040,
        1045, 1050, 1055, 1060, 1065, 1070, 1075, 1080, 1085, 1090, 1095, 1100, 1105, 1110, 1115, 1120, 1125, 1130,
        1135, 1140, 1145, 1150, 1155, 1160, 1165, 1170, 1175, 1180, 1185, 1190, 1195, 1200, 1205, 1210, 1215, 1220,
        1225, 1230, 1235, 1240, 1245, 1250, 1255, 1260, 1265, 1270, 1275, 1280, 1285, 1290, 1295, 1300, 1305, 1310,
        1315, 1320, 1325, 1330, 1335, 1340, 1345, 1350, 1355, 1360, 1365, 1370, 1375, 1380, 1385, 1390, 1395, 1400,
        1405, 1410, 1415, 1420, 1425, 1430, 1435, 1440, 1445, 1450, 1455, 1460, 1465, 1470, 1475, 1480, 1485, 1490,
        1495, 1500, 1505, 1510, 1515, 1520, 1525, 1530, 1535, 1540, 1545, 1550, 1555, 1560, 1565, 1570, 1575, 1580,
        1585, 1590, 1595, 1600, 1605, 1610, 1615, 1620, 1625, 1630, 1635, 1640, 1645, 1650, 1655, 1660, 1665, 1670,
        1675, 1680, 1685, 1690, 1695, 1700,
    )


@dataclass
class Meta:
    name: str = 'New profile'
    short_name_top: str = ''
    short_name_bot: str = ''
    user_note: str = ''


@dataclass
class Zeroing:
    x: float = 0
    y: float = 0
    pitch: float = 0
    distance: float = 100


@dataclass
class Atmosphere:
    temperature: int = 15
    pressure: int = 1000
    humidity: int = 50


@dataclass
class Barrel:
    caliber: str = 'New caliber'
    sight_height: int = 90
    twist: float = 9.
    twist_dir: profedit_pb2.TwistDir = profedit_pb2.TwistDir.RIGHT


@dataclass
class Cartridge:
    name: str = "New cartridge"
    muzzle_velocity: float = 800.
    temperature: int = 15
    powder_sens: float = 1.5


class DragPoint(NamedTuple):
    coeff: float
    velocity: float


@dataclass
class Bullet:
    name: str = "New bullet"
    diameter: float = 0.308
    weight: float = 178.
    length: float = 1.2
    drag_type: profedit_pb2.GType = profedit_pb2.GType.G7
    drag_model: tuple[DragPoint] = (DragPoint(1., 0.),)


class Switches(NamedTuple):
    s1: profedit_pb2.SwPos = profedit_pb2.SwPos(c_idx=255, zoom=1, distance=10000)
    s2: profedit_pb2.SwPos = profedit_pb2.SwPos(c_idx=255, zoom=2, distance=20000)
    s3: profedit_pb2.SwPos = profedit_pb2.SwPos(c_idx=255, zoom=3, distance=30000)
    s4: profedit_pb2.SwPos = profedit_pb2.SwPos(c_idx=255, zoom=4, distance=100000)


class A7PFactory:

    def __new__(cls,
                meta: Meta = Meta(),
                barrel: Barrel = Barrel(),
                cartridge: Cartridge = Cartridge(),
                bullet: Bullet = Bullet(),
                zeroing: Zeroing = Zeroing(),
                zero_atmo: Atmosphere = Atmosphere(),
                zero_powder_temp: int = 15,
                distances: [DistanceTable, tuple[float]] = DistanceTable.LONG_RANGE,
                switches: Switches = Switches()
                ) -> profedit_pb2.Payload:
        def fmt_bottom():
            if meta.short_name_bot:
                return meta.short_name_bot
            return "{:.{}f}".format(bullet.weight, 0 if bullet.weight % 1 == 0 else 1) + 'gr'

        def drag_model():
            return [
                profedit_pb2.CoefRow(
                    bc_cd=round(point.coeff * 10000),
                    mv=round(point.velocity * 10)
                ) for point in bullet.drag_model
            ]

        if isinstance(distances, DistanceTable):
            _distances = distances.value
        elif isinstance(distances, tuple):
            _distances = distances
        else:
            raise ValueError("Distances have to be an instance of DistanceTable or tuple[float]")
        if len(_distances) < 1:
            raise ValueError("List of distances can't be empty")
        zero_dist_idx = _distances.index(round(zeroing.distance))

        return profedit_pb2.Payload(
            profile=profedit_pb2.Profile(
                profile_name=meta.name,
                cartridge_name=cartridge.name,
                bullet_name=bullet.name,

                short_name_top=meta.short_name_top if meta.short_name_top else meta.name[:6],
                short_name_bot=meta.short_name_bot if meta.short_name_bot else fmt_bottom(),
                user_note=meta.user_note,

                zero_x=round(zeroing.x * -1000),  # click_x * -1000
                zero_y=round(zeroing.y * 1000),  # click_y * 1000
                sc_height=barrel.sight_height,  # mm
                r_twist=round(barrel.twist * 100),  # Inch * 100

                c_muzzle_velocity=round(cartridge.muzzle_velocity * 10),  # MPS * 10
                c_zero_temperature=cartridge.temperature,  # C
                c_t_coeff=round(cartridge.powder_sens * 1000),  # % * 1000
                c_zero_air_temperature=zero_atmo.temperature,
                c_zero_air_pressure=round(zero_atmo.pressure * 10),  # hPa * 10
                c_zero_air_humidity=zero_atmo.humidity,  # %
                c_zero_p_temperature=zero_powder_temp,  # C
                c_zero_w_pitch=round(zeroing.pitch * 10),  # degree * 10

                b_diameter=round(bullet.diameter * 1000),  # Inch * 1000
                b_weight=round(bullet.weight * 10),  # grain * 10
                b_length=round(bullet.length * 1000),  # Inch * 1000

                twist_dir=barrel.twist_dir,
                bc_type=bullet.drag_type,  # G1, G7, CUSTOM

                switches=list(switches),
                distances=[round(d * 100) for d in _distances],
                c_zero_distance_idx=zero_dist_idx if zero_dist_idx else 0,
                coef_rows=drag_model(),

                caliber=barrel.caliber
            )
        )

    Meta = Meta
    Barrel = Barrel
    Cartridge = Cartridge
    Bullet = Bullet
    Zeroing = Zeroing
    Atmosphere = Atmosphere
    DistanceTable = DistanceTable
    Switches = Switches
    DragPoint = DragPoint



# Copyright 2023 Buf Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from a7p.protovalidate import validator

Validator = validator.Validator
CompilationError = validator.CompilationError
ValidationError = validator.ValidationError
Violations = validator.Violations

_validator = Validator()
validate = _validator.validate
collect_violations = _validator.collect_violations

__all__ = ["Validator", "CompilationError", "ValidationError", "Violations", "validate", "collect_violations"]



from unittest import TestCase

from a7p import A7PFile, A7PDataError


class TestA7P(TestCase):

    def setUp(self) -> None:
        self.test_data = b'73e771793eefa7138da2ad842f97ffb5\n\xe2\x02\n\x0212\x12\x011\x1a\x011"\x011*\x0118\xe8\x07@' \
                    b'\xe8\x07H\x01PdXn`\x01h\xe8\x07x\x0f\x80\x01\x90N\x88\x01(\x98\x01\x0f\xa0\x01\xe0+\xa8\x01' \
                    b'\n\xb0\x01\xe8\x07\xca\x01\x08\x08\xff\x01\x18\x01 \x90N\xca\x01\t\x08\xff\x01\x18\x02 \xa0' \
                    b'\x9c\x01\xca\x01\t\x08\xff\x01\x18\x03 \xb0\xea\x01\xca\x01\t\x08\xff\x01\x18\x04 \xa0\x8d' \
                    b'\x06\xd2\x01\xe9\x01\x90N\xa0\x9c\x01\xa8\xc3\x01\xb0\xea\x01\xf4\xfd\x01\xb8\x91\x02\xfc\xa4' \
                    b'\x02\xc0\xb8\x02\x90\xc8\x02\xe0\xd7\x02\xb0\xe7\x02\x80\xf7\x02\xd0\x86\x03\xa0\x96\x03\xf0' \
                    b'\xa5\x03\xc0\xb5\x03\x90\xc5\x03\xe0\xd4\x03\xc8\xdc\x03\xb0\xe4\x03\x98\xec\x03\x80\xf4\x03' \
                    b'\xe8\xfb\x03\xd0\x83\x04\xb8\x8b\x04\xa0\x93\x04\x88\x9b\x04\xf0\xa2\x04\xd8\xaa\x04\xc0\xb2' \
                    b'\x04\xa8\xba\x04\x90\xc2\x04\xf8\xc9\x04\xe0\xd1\x04\xc8\xd9\x04\xb0\xe1\x04\x98\xe9\x04\x80' \
                    b'\xf1\x04\xf4\xf4\x04\xe8\xf8\x04\xdc\xfc\x04\xd0\x80\x05\xc4\x84\x05\xb8\x88\x05\xac\x8c\x05' \
                    b'\xa0\x90\x05\x94\x94\x05\x88\x98\x05\xfc\x9b\x05\xf0\x9f\x05\xe4\xa3\x05\xd8\xa7\x05\xcc\xab' \
                    b'\x05\xc0\xaf\x05\xb4\xb3\x05\xa8\xb7\x05\x9c\xbb\x05\x90\xbf\x05\x84\xc3\x05\xf8\xc6\x05\xec' \
                    b'\xca\x05\xe0\xce\x05\xd4\xd2\x05\xc8\xd6\x05\xbc\xda\x05\xb0\xde\x05\xa4\xe2\x05\x98\xe6\x05' \
                    b'\x8c\xea\x05\x80\xee\x05\xf4\xf1\x05\xe8\xf5\x05\xdc\xf9\x05\xd0\xfd\x05\xc4\x81\x06\xb8\x85' \
                    b'\x06\xac\x89\x06\xa0\x8d\x06\xda\x01\x03\x08\x90N\xe2\x01\x0522 LR'

    def testValidate(self):
        profile = A7PFile.loads(self.test_data)
        build = A7PFile.dumps(profile)
        self.assertEqual(self.test_data, build)

        json_ = A7PFile.to_json(profile)
        print(json_)
        print(A7PFile.from_json(json_))

        dict_ = A7PFile.to_dict(profile)
        print(dict_)
        print(A7PFile.from_dict(dict_))

    def testException(self):
        is_raise_exception = False
        try:
            A7PFile.loads(b'')
        except A7PDataError:
            is_raise_exception = True
        self.assertTrue(is_raise_exception, "A7PDataError exception didn't raised")




# Copyright 2023 Buf Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from google.protobuf import message

from a7p.buf.validate import expression_pb2  # type: ignore
from a7p.protovalidate.internal import constraints as _constraints
from a7p.protovalidate.internal import extra_func

CompilationError = _constraints.CompilationError
Violations = expression_pb2.Violations


class Validator:
    """
    Validates protobuf messages against static constraints.

    Each validator instance caches internal state generated from the static
    constraints, so reusing the same instance for multiple validations
    significantly improves performance.
    """

    _factory: _constraints.ConstraintFactory

    def __init__(self):
        self._factory = _constraints.ConstraintFactory(extra_func.EXTRA_FUNCS)

    def validate(
        self,
        message: message.Message,
        *,
        fail_fast: bool = False,
    ):
        """
        Validates the given message against the static constraints defined in
        the message's descriptor.

        Parameters:
            message: The message to validate.
            fail_fast: If true, validation will stop after the first violation.
        Raises:
            CompilationError: If the static constraints could not be compiled.
            ValidationError: If the message is invalid.
        """
        violations = self.collect_violations(message, fail_fast=fail_fast)
        if violations.violations:
            msg = f"invalid {message.DESCRIPTOR.name}"
            raise ValidationError(msg, violations)

    def collect_violations(
        self,
        message: message.Message,
        *,
        fail_fast: bool = False,
        into: expression_pb2.Violations = None,
    ) -> expression_pb2.Violations:
        """
        Validates the given message against the static constraints defined in
        the message's descriptor. Compared to validate, collect_violations is
        faster but puts the burden of raising an appropriate exception on the
        caller.

        Parameters:
            message: The message to validate.
            fail_fast: If true, validation will stop after the first violation.
            into: If provided, any violations will be appended to the
                Violations object and the same object will be returned.
        Raises:
            CompilationError: If the static constraints could not be compiled.
        """
        ctx = _constraints.ConstraintContext(fail_fast=fail_fast, violations=into)
        for constraint in self._factory.get(message.DESCRIPTOR):
            constraint.validate(ctx, message)
            if ctx.done:
                break
        return ctx.violations


class ValidationError(ValueError):
    """
    An error raised when a message fails to validate.
    """

    violations: expression_pb2.Violations

    def __init__(self, msg: str, violations: expression_pb2.Violations):
        super().__init__(msg)
        self.violations = violations

    def errors(self) -> typing.List[expression_pb2.Violation]:
        """
        Returns the validation errors as a simple Python list, rather than the
        Protobuf-specific collection type used by Violations.
        """
        return list(self.violations.violations)



# Copyright 2023 Buf Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing

import celpy  # type: ignore
from celpy import celtypes  # type: ignore
from google.protobuf import any_pb2, descriptor, message

from a7p.buf.validate import expression_pb2, validate_pb2  # type: ignore
from a7p.buf.validate.priv import private_pb2  # type: ignore
from a7p.protovalidate.internal import string_format


class CompilationError(Exception):
    pass


def make_key_path(field_name: str, key: celtypes.Value) -> str:
    return f"{field_name}[{string_format.format_value(key)}]"


def make_duration(msg: message.Message) -> celtypes.DurationType:
    return celtypes.DurationType(
        seconds=msg.seconds,  # type: ignore
        nanos=msg.nanos,  # type: ignore
    )


def make_timestamp(msg: message.Message) -> celtypes.TimestampType:
    return make_duration(msg) + celtypes.TimestampType(1970, 1, 1)


def unwrap(msg: message.Message) -> celtypes.Value:
    return _field_to_cel(msg, msg.DESCRIPTOR.fields_by_name["value"])


_MSG_TYPE_URL_TO_CTOR = {
    "google.protobuf.Duration": make_duration,
    "google.protobuf.Timestamp": make_timestamp,
    "google.protobuf.StringValue": unwrap,
    "google.protobuf.BytesValue": unwrap,
    "google.protobuf.Int32Value": unwrap,
    "google.protobuf.Int64Value": unwrap,
    "google.protobuf.UInt32Value": unwrap,
    "google.protobuf.UInt64Value": unwrap,
    "google.protobuf.FloatValue": unwrap,
    "google.protobuf.DoubleValue": unwrap,
    "google.protobuf.BoolValue": unwrap,
}


def _msg_to_cel(msg: message.Message) -> dict[str, celtypes.Value]:
    ctor = _MSG_TYPE_URL_TO_CTOR.get(msg.DESCRIPTOR.full_name)
    if ctor is not None:
        return ctor(msg)
    result = celtypes.MapType()
    field: descriptor.FieldDescriptor
    for field in msg.DESCRIPTOR.fields:
        if field.containing_oneof is not None and not msg.HasField(field.name):
            continue
        result[field.name] = _field_to_cel(msg, field)
    return result


_TYPE_TO_CTOR = {
    descriptor.FieldDescriptor.TYPE_MESSAGE: _msg_to_cel,
    descriptor.FieldDescriptor.TYPE_ENUM: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_BOOL: celtypes.BoolType,
    descriptor.FieldDescriptor.TYPE_BYTES: celtypes.BytesType,
    descriptor.FieldDescriptor.TYPE_STRING: celtypes.StringType,
    descriptor.FieldDescriptor.TYPE_FLOAT: celtypes.DoubleType,
    descriptor.FieldDescriptor.TYPE_DOUBLE: celtypes.DoubleType,
    descriptor.FieldDescriptor.TYPE_INT32: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_INT64: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_UINT32: celtypes.UintType,
    descriptor.FieldDescriptor.TYPE_UINT64: celtypes.UintType,
    descriptor.FieldDescriptor.TYPE_SINT32: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_SINT64: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_FIXED32: celtypes.UintType,
    descriptor.FieldDescriptor.TYPE_FIXED64: celtypes.UintType,
    descriptor.FieldDescriptor.TYPE_SFIXED32: celtypes.IntType,
    descriptor.FieldDescriptor.TYPE_SFIXED64: celtypes.IntType,
}


def _scalar_field_value_to_cel(val: typing.Any, field: descriptor.FieldDescriptor) -> celtypes.Value:
    ctor = _TYPE_TO_CTOR.get(field.type)
    if ctor is None:
        msg = "unknown field type"
        raise CompilationError(msg)
    return ctor(val)


def _field_value_to_cel(val: typing.Any, field: descriptor.FieldDescriptor) -> celtypes.Value:
    if field.label == descriptor.FieldDescriptor.LABEL_REPEATED:
        if field.message_type is not None and field.message_type.GetOptions().map_entry:
            return _map_field_value_to_cel(val, field)
        return _repeated_field_value_to_cel(val, field)
    return _scalar_field_value_to_cel(val, field)


def _is_empty_field(msg: message.Message, field: descriptor.FieldDescriptor) -> bool:
    if field.containing_oneof is not None and not msg.HasField(field.name):
        return True
    if field.label == descriptor.FieldDescriptor.LABEL_REPEATED:
        return len(getattr(msg, field.name)) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_MESSAGE:
        return not msg.HasField(field.name)
    if field.type == descriptor.FieldDescriptor.TYPE_BOOL:
        return not getattr(msg, field.name)
    if field.type == descriptor.FieldDescriptor.TYPE_BYTES:
        return len(getattr(msg, field.name)) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_STRING:
        return len(getattr(msg, field.name)) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_FLOAT:
        return getattr(msg, field.name) == 0.0
    if field.type == descriptor.FieldDescriptor.TYPE_DOUBLE:
        return getattr(msg, field.name) == 0.0
    if field.type == descriptor.FieldDescriptor.TYPE_INT32:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_INT64:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_UINT32:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_UINT64:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_SINT32:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_SINT64:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_FIXED32:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_FIXED64:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_SFIXED32:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_SFIXED64:
        return getattr(msg, field.name) == 0
    if field.type == descriptor.FieldDescriptor.TYPE_ENUM:
        return getattr(msg, field.name) == 0
    exception_msg = "unknown field type"
    raise ValueError(exception_msg)


def _repeated_field_to_cel(msg: message.Message, field: descriptor.FieldDescriptor) -> celtypes.Value:
    if field.message_type is not None and field.message_type.GetOptions().map_entry:
        return _map_field_to_cel(msg, field)
    return _repeated_field_value_to_cel(getattr(msg, field.name), field)


def _repeated_field_value_to_cel(val: typing.Any, field: descriptor.FieldDescriptor) -> celtypes.Value:
    result = celtypes.ListType()
    for item in val:
        result.append(_scalar_field_value_to_cel(item, field))
    return result


def _map_field_value_to_cel(mapping: typing.Any, field: descriptor.FieldDescriptor) -> celtypes.Value:
    result = celtypes.MapType()
    key_field = field.message_type.fields[0]
    val_field = field.message_type.fields[1]
    for key, val in mapping.items():
        result[_field_value_to_cel(key, key_field)] = _field_value_to_cel(val, val_field)
    return result


def _map_field_to_cel(msg: message.Message, field: descriptor.FieldDescriptor) -> celtypes.Value:
    return _map_field_value_to_cel(getattr(msg, field.name), field)


def _field_to_cel(msg: message.Message, field: descriptor.FieldDescriptor) -> celtypes.Value:
    if field.label == descriptor.FieldDescriptor.LABEL_REPEATED:
        return _repeated_field_to_cel(msg, field)
    elif field.message_type is not None and not msg.HasField(field.name):
        return None
    else:
        return _scalar_field_value_to_cel(getattr(msg, field.name), field)


class ConstraintContext:
    """The state associated with a single constraint evaluation."""

    def __init__(self, fail_fast: bool = False, violations: expression_pb2.Violations = None):  # noqa: FBT001, FBT002
        self._fail_fast = fail_fast
        if violations is None:
            violations = expression_pb2.Violations()
        self._violations = violations

    @property
    def fail_fast(self) -> bool:
        return self._fail_fast

    @property
    def violations(self) -> expression_pb2.Violations:
        return self._violations

    def add(self, field_name: str, constraint_id: str, message: str, *, for_key: bool = False):
        self._violations.violations.append(
            expression_pb2.Violation(
                field_path=field_name,
                constraint_id=constraint_id,
                message=message,
                for_key=for_key,
            )
        )

    def add_errors(self, other_ctx):
        self._violations.violations.extend(other_ctx.violations.violations)

    def add_path_prefix(self, prefix: str, delim="."):
        for violation in self._violations.violations:
            if violation.field_path:
                violation.field_path = prefix + delim + violation.field_path
            else:
                violation.field_path = prefix

    @property
    def done(self) -> bool:
        return self._fail_fast and self.has_errors()

    def has_errors(self) -> bool:
        return len(self._violations.violations) > 0

    def sub_context(self):
        return ConstraintContext(self._fail_fast)


class ConstraintRules:
    """The constraints associated with a single 'rules' message."""

    def validate(self, ctx: ConstraintContext, message: message.Message):  # noqa: ARG002
        """Validate the message against the rules in this constraint."""
        ctx.add("", "unimplemented", "Unimplemented")


class CelConstraintRules(ConstraintRules):
    """A constraint that has rules written in CEL."""

    _runners: list[typing.Tuple[celpy.Runner, expression_pb2.Constraint | private_pb2.Constraint]]
    _rules_cel: celtypes.Value = None

    def __init__(self, rules: message.Message | None):
        self._runners = []
        if rules is not None:
            self._rules_cel = _msg_to_cel(rules)

    def _validate_cel(
        self, ctx: ConstraintContext, field_name: str, activation: dict[str, typing.Any], *, for_key: bool = False
    ):
        activation["rules"] = self._rules_cel
        activation["now"] = celtypes.TimestampType(datetime.datetime.now(tz=datetime.timezone.utc))
        for runner, constraint in self._runners:
            result = runner.evaluate(activation)
            if isinstance(result, celtypes.BoolType):
                if not result:
                    ctx.add(field_name, constraint.id, constraint.message, for_key=for_key)
            elif isinstance(result, celtypes.StringType):
                if result:
                    ctx.add(field_name, constraint.id, result, for_key=for_key)
            elif isinstance(result, Exception):
                raise result

    def add_rule(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        rules: expression_pb2.Constraint | private_pb2.Constraint,
    ):
        ast = env.compile(rules.expression)
        prog = env.program(ast, functions=funcs)
        self._runners.append((prog, rules))


class MessageConstraintRules(CelConstraintRules):
    """Message-level rules."""

    def validate(self, ctx: ConstraintContext, message: message.Message):
        self._validate_cel(ctx, "", {"this": _msg_to_cel(message)})


def check_field_type(field: descriptor.FieldDescriptor, expected: int, wrapper_name: str | None = None):
    if field.type != expected and (
        field.type != descriptor.FieldDescriptor.TYPE_MESSAGE or field.message_type.full_name != wrapper_name
    ):
        msg = f"field {field.name} has type {field.type} but expected {expected}"
        raise CompilationError(msg)


class FieldConstraintRules(CelConstraintRules):
    """Field-level rules."""

    _ignore_empty = False
    _required = False

    def __init__(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.FieldConstraints,
    ):
        type_case = field_level.WhichOneof("type")
        super().__init__(None if type_case is None else getattr(field_level, type_case))
        self._field = field
        if field_level.ignore_empty:
            self._ignore_empty = True
        if field_level.required:
            self._required = True
        type_case = field_level.WhichOneof("type")
        if type_case is not None:
            rules = getattr(field_level, type_case)
            # For each set field in the message, look for the private constraint
            # extension.
            for field, _ in rules.ListFields():
                if private_pb2.field in field.GetOptions().Extensions:
                    for cel in field.GetOptions().Extensions[private_pb2.field].cel:
                        self.add_rule(env, funcs, cel)
        for cel in field_level.cel:
            self.add_rule(env, funcs, cel)

    def validate(self, ctx: ConstraintContext, message: message.Message):
        if _is_empty_field(message, self._field):
            if self._required:
                ctx.add(
                    self._field.name,
                    "required",
                    "value is required",
                )
                return
            if (
                self._ignore_empty
                or (
                    self._field.label != descriptor.FieldDescriptor.LABEL_REPEATED
                    and self._field.type == descriptor.FieldDescriptor.TYPE_MESSAGE
                )
                or self._field.containing_oneof is not None
            ):
                return
        val = getattr(message, self._field.name)
        self._validate_value(ctx, self._field.name, val)
        self._validate_cel(ctx, self._field.name, {"this": _field_value_to_cel(val, self._field)})

    def validate_item(self, ctx: ConstraintContext, field_path: str, val: typing.Any, *, for_key: bool = False):
        self._validate_value(ctx, field_path, val, for_key=for_key)
        self._validate_cel(ctx, field_path, {"this": _scalar_field_value_to_cel(val, self._field)}, for_key=for_key)

    def _validate_value(self, ctx: ConstraintContext, field_path: str, val: typing.Any, *, for_key: bool = False):
        pass


class AnyConstraintRules(FieldConstraintRules):
    """Rules for an Any field."""

    _in: typing.List[str] = []  # noqa: RUF012
    _not_in: typing.List[str] = []  # noqa: RUF012

    def __init__(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.FieldConstraints,
    ):
        super().__init__(env, funcs, field, field_level)
        if getattr(field_level.any, "in"):
            self._in = getattr(field_level.any, "in")
        if field_level.any.not_in:
            self._not_in = field_level.any.not_in

    def _validate_value(self, ctx: ConstraintContext, field_path: str, value: any_pb2.Any, *, for_key: bool = False):
        if len(self._in) > 0:
            if value.type_url not in self._in:
                ctx.add(
                    field_path,
                    "any.in",
                    "type URL must be in the allow list",
                    for_key=for_key,
                )
        if value.type_url in self._not_in:
            ctx.add(
                field_path,
                "any.not_in",
                "type URL must not be in the block list",
                for_key=for_key,
            )


class EnumConstraintRules(FieldConstraintRules):
    """Rules for an enum field."""

    _defined_only = False

    def __init__(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.FieldConstraints,
    ):
        super().__init__(env, funcs, field, field_level)
        if field_level.enum.defined_only:
            self._defined_only = True

    def validate(self, ctx: ConstraintContext, message: message.Message):
        super().validate(ctx, message)
        if ctx.done:
            return
        if self._defined_only:
            value = getattr(message, self._field.name)
            if value not in self._field.enum_type.values_by_number:
                ctx.add(
                    self._field.name,
                    "enum.defined_only",
                    "value must be one of the defined enum values",
                )


class RepeatedConstraintRules(FieldConstraintRules):
    """Rules for a repeated field."""

    _item_rules: FieldConstraintRules | None = None

    def __init__(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.FieldConstraints,
        item_rules: FieldConstraintRules | None,
    ):
        super().__init__(env, funcs, field, field_level)
        if item_rules is not None:
            self._item_rules = item_rules

    def validate(self, ctx: ConstraintContext, message: message.Message):
        super().validate(ctx, message)
        if ctx.done:
            return
        value = getattr(message, self._field.name)
        if self._item_rules is not None:
            for i, item in enumerate(value):
                sub_ctx = ctx.sub_context()
                self._item_rules.validate_item(sub_ctx, "", item)
                if sub_ctx.has_errors():
                    sub_ctx.add_path_prefix(f"{self._field.name}[{i}]", "")
                    ctx.add_errors(sub_ctx)
                if ctx.done:
                    return


class MapConstraintRules(FieldConstraintRules):
    """Rules for a map field."""

    _key_rules: FieldConstraintRules | None = None
    _value_rules: FieldConstraintRules | None = None

    def __init__(
        self,
        env: celpy.Environment,
        funcs: dict[str, celpy.CELFunction],
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.FieldConstraints,
        key_rules: FieldConstraintRules | None,
        value_rules: FieldConstraintRules | None,
    ):
        super().__init__(env, funcs, field, field_level)
        if key_rules is not None:
            self._key_rules = key_rules
        if value_rules is not None:
            self._value_rules = value_rules

    def validate(self, ctx: ConstraintContext, message: message.Message):
        super().validate(ctx, message)
        if ctx.done:
            return
        value = getattr(message, self._field.name)
        for k, v in value.items():
            key_field_path = make_key_path(self._field.name, k)
            if self._key_rules is not None:
                self._key_rules.validate_item(ctx, key_field_path, k, for_key=True)
            if self._value_rules is not None:
                self._value_rules.validate_item(ctx, key_field_path, v)


class OneofConstraintRules(ConstraintRules):
    """Rules for a oneof definition."""

    required = True

    def __init__(self, oneof: descriptor.OneofDescriptor, rules: validate_pb2.OneofConstraints):
        self._oneof = oneof
        if not rules.required:
            self.required = False

    def validate(self, ctx: ConstraintContext, message: message.Message):
        if not message.WhichOneof(self._oneof.name):
            if self.required:
                ctx.add(
                    self._oneof.name,
                    "required",
                    "exactly one field is required in oneof",
                )
            return


class ConstraintFactory:
    """Factory for creating and caching constraints."""

    _env: celpy.Environment
    _funcs: dict[str, celpy.CELFunction]
    _cache: dict[descriptor.Descriptor, list[ConstraintRules] | Exception]

    def __init__(self, funcs: dict[str, celpy.CELFunction]):
        self._env = celpy.Environment()
        self._funcs = funcs
        self._cache = {}

    def get(self, descriptor: descriptor.Descriptor) -> list[ConstraintRules]:
        if descriptor not in self._cache:
            try:
                self._cache[descriptor] = self._new_constraints(descriptor)
            except Exception as e:
                self._cache[descriptor] = e
        result = self._cache[descriptor]
        if isinstance(result, Exception):
            raise result
        return result

    def _new_message_constraint(self, rules: validate_pb2.message) -> MessageConstraintRules:
        result = MessageConstraintRules(rules)
        for cel in rules.cel:
            result.add_rule(self._env, self._funcs, cel)
        return result

    def _new_scalar_field_constraint(
        self,
        field: descriptor.FieldDescriptor,
        field_level: validate_pb2.field,
    ):
        if field_level.skipped:
            return None
        type_case = field_level.WhichOneof("type")
        if type_case is None:
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "duration":
            check_field_type(field, 0, "google.protobuf.Duration")
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "timestamp":
            check_field_type(field, 0, "google.protobuf.Timestamp")
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "enum":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_ENUM)
            result = EnumConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "bool":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_BOOL, "google.protobuf.BoolValue")
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "bytes":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_BYTES,
                "google.protobuf.BytesValue",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "fixed32":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_FIXED32)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "fixed64":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_FIXED64)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "float":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_FLOAT,
                "google.protobuf.FloatValue",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "double":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_DOUBLE,
                "google.protobuf.DoubleValue",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "int32":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_INT32,
                "google.protobuf.Int32Value",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "int64":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_INT64,
                "google.protobuf.Int64Value",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "sfixed32":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_SFIXED32)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "sfixed64":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_SFIXED64)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "sint32":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_SINT32)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "sint64":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_SINT64)
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "uint32":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_UINT32,
                "google.protobuf.UInt32Value",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "uint64":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_UINT64,
                "google.protobuf.UInt64Value",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "string":
            check_field_type(
                field,
                descriptor.FieldDescriptor.TYPE_STRING,
                "google.protobuf.StringValue",
            )
            result = FieldConstraintRules(self._env, self._funcs, field, field_level)
            return result
        elif type_case == "any":
            check_field_type(field, descriptor.FieldDescriptor.TYPE_MESSAGE, "google.protobuf.Any")
            result = AnyConstraintRules(self._env, self._funcs, field, field_level)
            return result

    def _new_field_constraint(
        self,
        field: descriptor.FieldDescriptor,
        rules: validate_pb2.field,
    ) -> FieldConstraintRules:
        if field.label != descriptor.FieldDescriptor.LABEL_REPEATED:
            return self._new_scalar_field_constraint(field, rules)
        if field.message_type is not None and field.message_type.GetOptions().map_entry:
            key_rules = None
            if rules.map.HasField("keys"):
                key_field = field.message_type.fields_by_name["key"]
                key_rules = self._new_scalar_field_constraint(key_field, rules.map.keys)
            value_rules = None
            if rules.map.HasField("values"):
                value_field = field.message_type.fields_by_name["value"]
                value_rules = self._new_scalar_field_constraint(value_field, rules.map.values)
            return MapConstraintRules(self._env, self._funcs, field, rules, key_rules, value_rules)
        item_rule = None
        if rules.repeated.HasField("items"):
            item_rule = self._new_scalar_field_constraint(field, rules.repeated.items)
        return RepeatedConstraintRules(self._env, self._funcs, field, rules, item_rule)

    def _new_constraints(self, desc: descriptor.Descriptor) -> list[ConstraintRules]:
        result: list[ConstraintRules] = []
        constraint: ConstraintRules | None = None
        if validate_pb2.message in desc.GetOptions().Extensions:
            message_level = desc.GetOptions().Extensions[validate_pb2.message]
            if message_level.disabled:
                return []
            if constraint := self._new_message_constraint(message_level):
                result.append(constraint)

        for oneof in desc.oneofs:
            if validate_pb2.oneof in oneof.GetOptions().Extensions:
                if constraint := OneofConstraintRules(oneof, oneof.GetOptions().Extensions[validate_pb2.oneof]):
                    result.append(constraint)

        for field in desc.fields:
            if validate_pb2.field in field.GetOptions().Extensions:
                field_level = field.GetOptions().Extensions[validate_pb2.field]
                if field_level.skipped:
                    continue
                result.append(self._new_field_constraint(field, field_level))
                if field_level.repeated.items.skipped:
                    continue
            if field.message_type is None:
                continue
            if field.message_type.GetOptions().map_entry:
                value_field = field.message_type.fields_by_name["value"]
                if value_field.type != descriptor.FieldDescriptor.TYPE_MESSAGE:
                    continue
                result.append(MapValMsgConstraint(self, field, value_field))
            elif field.label == descriptor.FieldDescriptor.LABEL_REPEATED:
                result.append(RepeatedMsgConstraint(self, field))
            else:
                result.append(SubMsgConstraint(self, field))
        return result


class SubMsgConstraint(ConstraintRules):
    def __init__(
        self,
        factory: ConstraintFactory,
        field: descriptor.FieldDescriptor,
    ):
        self._factory = factory
        self._field = field

    def validate(self, ctx: ConstraintContext, message: message.Message):
        if not message.HasField(self._field.name):
            return
        constraints = self._factory.get(self._field.message_type)
        if constraints is None:
            return
        val = getattr(message, self._field.name)
        sub_ctx = ctx.sub_context()
        for constraint in constraints:
            constraint.validate(sub_ctx, val)
        if sub_ctx.has_errors():
            sub_ctx.add_path_prefix(self._field.name)
            ctx.add_errors(sub_ctx)


class MapValMsgConstraint(ConstraintRules):
    def __init__(
        self,
        factory: ConstraintFactory,
        field: descriptor.FieldDescriptor,
        value_field: descriptor.FieldDescriptor,
    ):
        self._factory = factory
        self._field = field
        self._value_field = value_field

    def validate(self, ctx: ConstraintContext, message: message.Message):
        val = getattr(message, self._field.name)
        if not val:
            return
        constraints = self._factory.get(self._value_field.message_type)
        if constraints is None:
            return
        for k, v in val.items():
            sub_ctx = ctx.sub_context()
            for constraint in constraints:
                constraint.validate(sub_ctx, v)
            if sub_ctx.has_errors():
                sub_ctx.add_path_prefix(f"{self._field.name}[{k}]")
                ctx.add_errors(sub_ctx)


class RepeatedMsgConstraint(ConstraintRules):
    def __init__(
        self,
        factory: ConstraintFactory,
        field: descriptor.FieldDescriptor,
    ):
        self._factory = factory
        self._field = field

    def validate(self, ctx: ConstraintContext, message: message.Message):
        val = getattr(message, self._field.name)
        if not val:
            return
        constraints = self._factory.get(self._field.message_type)
        if constraints is None:
            return
        for idx, item in enumerate(val):
            sub_ctx = ctx.sub_context()
            for constraint in constraints:
                constraint.validate(sub_ctx, item)
            if sub_ctx.has_errors():
                sub_ctx.add_path_prefix(f"{self._field.name}[{idx}]")
                ctx.add_errors(sub_ctx)



import hashlib
import json
from typing import BinaryIO

from a7p import profedit_pb2
from google.protobuf.json_format import MessageToJson, MessageToDict, Parse
from a7p import protovalidate as validator

__all__ = ['A7PFile', 'A7PDataError']


class A7PDataError(Exception):
    pass


class A7PFile:

    @staticmethod
    def loads(string: bytes, validate: bool = True):
        data = string[32:]
        md5_hash = hashlib.md5(data).hexdigest()
        if md5_hash == string[:32].decode():
            profile = profedit_pb2.Payload()
            profile.ParseFromString(data)
            if validate:
                validator.validate(profile)
            return profile
        else:
            raise A7PDataError("Input data is missing for MD5 hashing")

    @staticmethod
    def load(file: BinaryIO, validate: bool = True) -> profedit_pb2.Payload:
        string = file.read()
        return A7PFile.loads(string, validate)

    @staticmethod
    def dumps(profile: profedit_pb2.Payload, validate: bool = True) -> bytes:
        if validate:
            validator.validate(profile)
        data = profile.SerializeToString()
        md5_hash = hashlib.md5(data).hexdigest().encode()
        return md5_hash + data

    @staticmethod
    def dump(profile: profedit_pb2.Payload, file: BinaryIO, validate: bool = True) -> None:
        data = A7PFile.dumps(profile, validate)
        file.write(data)

    @staticmethod
    def to_json(profile: profedit_pb2.Payload) -> str:
        return MessageToJson(profile)

    @staticmethod
    def from_json(json_data: str) -> profedit_pb2.Payload:
        return Parse(json_data, profedit_pb2.Payload())

    @staticmethod
    def to_dict(profile: profedit_pb2.Payload) -> dict:
        return MessageToDict(profile)

    @staticmethod
    def from_dict(data: dict) -> profedit_pb2.Payload:
        return Parse(json.dumps(data), profedit_pb2.Payload())


if __name__ == '__main__':
    with open('test.a7p', 'rb') as a7p:
        a7p_obj = A7PFile.load(a7p)
        a7p_dict = A7PFile.to_dict(a7p_obj)
        a7p_json = A7PFile.to_json(a7p_obj)
        a7p_by_dict = A7PFile.from_dict(a7p_dict)
        a7p_by_json = A7PFile.from_json(a7p_json)



import asyncio
import pathlib
import sys
from argparse import ArgumentParser
from dataclasses import dataclass
from importlib import metadata
from asyncio import Semaphore

from a7p.a7p import A7PFile, A7PDataError
from a7p import protovalidate
from a7p.protovalidate import ValidationError, Violations

try:
    __version__ = metadata.version("a7p")
except metadata.PackageNotFoundError:
    __version__ = "undefined version"

# Define a global Semaphore with a maximum number of threads
MAX_THREADS = 5  # Set the maximum number of threads
semaphore = Semaphore(MAX_THREADS)


async def limited_to_thread(func, *args):
    """Run a function in a thread, limited by the semaphore."""
    async with semaphore:  # Acquire the semaphore
        return await asyncio.to_thread(func, *args)


class CustomArgumentParser(ArgumentParser):
    def error(self, message):
        """Override error method to show help message on argument errors."""
        self.print_help(sys.stderr)
        sys.stderr.write(f"\nError: {message}\n")
        sys.exit(2)


parser = CustomArgumentParser(
    f"a7p {__version__}",
    exit_on_error=True,
)
parser.add_argument("path", type=pathlib.Path, help="Path to the directory or file")
parser.add_argument('-V', '--version', action='version', version=__version__)
parser.add_argument('-r', '--recursive', action='store_true', help="Recursively walk files")
parser.add_argument('-v', '--validate', action='store_true', help="Validate files")
parser.add_argument('--verbose', action='store_true', help="Verbose")

# parser.add_argument('-d', '--distances', action='store', help="Set distances set",
#                     choices=['subsonic', 'medium', 'long', 'ultra'])
# parser.add_argument('-zs', '--zero-sync', action='store_true', help="Synchronize zero")
# parser.add_argument('-zo', '--zero-offset', action='store', nargs=2, help="Set clicks offset",
#                     metavar=("X_OFFSET", "Y_OFFSET"))
# parser.add_argument('-cs', '--clicks-switch', action='store', nargs=4, help="Switch clicks sizes",
#                     metavar=("CUR_X", "NEW_X", "CUR_Y", "NEW_Y"))
#
# parser.add_argument('--json', action='store', type=pathlib.Path, help="Convert to/from JSON")
#
# parser.add_argument('--max-threads', action='store', type=int, default=5)


@dataclass
class Result:
    path: pathlib.Path
    error = None
    violations: Violations = None
    zero: tuple[float, float] = None

    def print(self):
        print(f"File: {self.path.absolute()}")
        if self.zero:
            print("Zero: X: {}, Y: {}".format(*self.zero))
        if self.error:
            print(f"Error: {self.error}")

    def details(self):

        if self.violations:
            for violation in self.violations.ListFields():
                try:
                    violation_msg = []
                    for details in violation[1]:
                        violation_msg.extend([
                            "\tViolation:"
                            f"\t\t{details.field_path}",
                            f"\t\t{details.constraint_id}",
                            f"\t\t{details.message}"
                        ])
                    print("\n".join(violation_msg))
                except Exception:
                    print(violation)


def process_file(path, validate=False, distances=None):
    if path.suffix != ".a7p":
        return
    result = Result(path)
    try:
        with open(path, 'rb') as fp:
            data = fp.read()
        payload = A7PFile.loads(data, False)

        try:
            if validate:
                protovalidate.validate(payload)
        except ValidationError as e:
            result.error = "Validation error"
            result.violations = e.violations
    except (IOError, A7PDataError) as e:
        result.error = e
        return result
    result.zero = (payload.profile.zero_x / 1000, payload.profile.zero_y / 1000)
    return result


async def process_files(
        path: pathlib.Path = None,
        recursive: bool = False,
        validate: bool = False,
        distances: str = None,
        json: pathlib.Path = None,
        verbose: bool = False,
):
    tasks = []

    if not path.is_dir():
        tasks.append(asyncio.to_thread(process_file, path, validate, distances))
    else:
        if json is not None:
            parser.error("--json conversion available only for a single file")
        if recursive:
            item: pathlib.Path
            for item in path.rglob("*"):  # '*' matches all files and directories
                if item.is_file():
                    tasks.append(limited_to_thread(process_file, item, validate, distances))
        else:
            for item in path.iterdir():
                if item.is_file():
                    tasks.append(limited_to_thread(process_file, item, validate, distances))

    results: list[Result] = await asyncio.gather(*tasks)

    for result in results:
        if result:
            result.print()
            if verbose:
                result.details()
            print()


def main():
    args = parser.parse_args()
    asyncio.run(process_files(**args.__dict__))


if __name__ == '__main__':
    main()



__author__ = "o-murphy"
__credits__ = ["Dmytro Yaroshenko"]
__copyright__ = ("",)

try:
    from a7p import profedit_pb2, factory
    from a7p import protovalidate as validator
    from a7p.a7p import A7PFile, A7PDataError
    from a7p.profedit_pb2 import *
    from a7p.factory import A7PFactory
except ImportError:
    import logging
    logging.warning("Can't import 'google' or 'protobuf' package, pass it on setup")

__all__ = ['A7PFile', 'A7PDataError', 'A7PFactory', 'factory',
           'profedit_pb2',
           'Payload', 'Profile', 'DType', 'GType', 'TwistDir', 'SwPos', 'CoefRow']



from setuptools import setup, find_packages

setup()
# from a7p import __version__, __author__
#
#
# CLASSIFIERS = [
#     'Intended Audience :: Developers',
#     'Natural Language :: English',
#     'Programming Language :: Python',
#     'Topic :: Software Development :: Libraries :: Python Modules'
# ]
#
# KEYWORDS = 'protobuf, archer, a7p'
#
#
# with open('requirements.txt', 'r') as fp:
#     setup_requires = fp.readlines()
#
#
# with open('README.MD', 'r') as fp:
#     long_description = fp.read()
#
#
# package_data = {'a7p': ['requirements.txt', 'a7p/test.a7p']}
#
# setup(
#     name='a7p',
#     version=__version__,
#     python_requires='>=3.9',
#
#     description='Simple python3 wrapper for .a7p files',
#     long_description=long_description,
#     long_description_content_type="text/markdown",
#
#     author=__author__,
#     url='https://github.com/o-murphy/a7p_transfer_example',
#     download_url='https://github.com/o-murphy/a7p_transfer_example',
#
#     classifiers=CLASSIFIERS,
#     keywords=KEYWORDS,
#
#     packages=find_packages(),
#     install_requires=setup_requires,
#     py_modules=['a7p'],
#
#     include_package_data=True,
#     package_data=package_data,
#
#     zip_safe=False,
# )

